#! /usr/bin/python

"""
The addressbook holds a list of permission slips for recipients to which we
are allowed to send mail, and incoming permissions for senders from whom we
will accept mail.

Each OwnerEntry has a separate AddressBook.
"""

from petmail import util, permission, transport, crypto, pbutil, request
from petmail.interfaces import IAddressBook, IIDRecord, IAddressBookEntry, \
     IOwnerEntry, IOwnerPermission, IPermissionSlip
from petmail.message import Message
from petmail.idrecord import IDRecord

from twisted.python.components import implements
from twisted.python import log
from twisted.internet import defer

import os

class AddressBook(object):
    # TODO: needs test coverage
    __implements__ = IAddressBook
    debug = 0

    def __init__(self, owner):
        assert implements(owner, IOwnerEntry)
        self.owner = owner
        self.entries = []
        self.watchers = []

    def __getstate__(self):
        d = self.__dict__.copy()
        d['watchers'] = []
        return d

    def notifyOnChange(self, callback):
        """Register a callback to be run when the addressbook changes (such
        as when an entry is added, perhaps due to receipt of a message from
        a new sender). The callback is run with one argument, this
        AddressBook instance. The callback is not persisted."""
        self.watchers.append(callback)
    def dontNotifyOnChange(self, callback):
        self.watchers.remove(callback)
    def notifyWatchers(self):
        for w in self.watchers:
            w(self)

    def lookupID(self, ID):
        if not isinstance(ID, type("")):
            print "AddressBook.lookupID got %s (type %s), not string" % \
                  (ID, type(ID))
            assert False
        for e in self.entries:
            if e.idRecord and e.idRecord.ID == ID:
                return e
        return None
    def lookupPetName(self, name):
        for e in self.entries:
            if e.petname == name:
                return e
        return None
    def lookupAddress(self, address):
        for e in self.entries:
            if address in e.idRecord.addresses:
                return e

    def addEntry(self, entry, message=None):
        # message is the message which caused this entry to be added
        # TODO: stash (parts of) the message
        assert(isinstance(entry, AddressBookEntry))
        entry.addressbook = self
        self.entries.append(entry)
        self.notifyWatchers()

    def addFromSigned(self, signed, reason="AddressBook.addFromSigned"):
        """Add the record to the address book, or update the existing entry.
        Return a Deferred which will fire (with the new entry) when the entry
        has been fully added.
        """

        e = AddressBookEntry()
        d = e.createFromIDRecord(signed, self.owner.parent.crypto, reason)
        d.addCallback(self._addFromSigned1, reason)
        return d

    def _addFromSigned1(self, entry, reason):
        # TODO: should addEntry accept a Message or simple text?
        self.addEntry(entry, reason)
        self.notifyWatchers()
        return entry


    def deleteEntry(self, entry):
        """Returns a Deferred which fires when the entry has been deleted.
        """
        if entry.theirPermission:
            entry.revokeTheirPermission()
            # would normally do entry.modified() here, but we need to push
            # out the update immediately instead of waiting for the holdoff
            # time
            d = entry.refreshRecipient()
        else:
            d = defer.succeed(None)
        d.addCallback(self._deleteEntry2, entry)
        return d

    def _deleteEntry2(self, results, entry):
        entry.addressbook = None
        self.entries.remove(entry)
        self.notifyWatchers()

class AddressBookEntry(object):
    __implements__ = IAddressBookEntry
    updateHoldoffTime = 30
    doUpdate = True

    def __init__(self):
        self.addressbook = None # set by AddressBook.addEntry
        self.petname = None
        self.idRecord = None
        self.ourPermission = None   # what we use to send mail to them
        # .ourPermission is always a PermissionSlip
        self.theirPermission = None # what we require when they send to us
        # .theirPermission is always an OwnerPermission
        self.vouchFor = True
        self.doNotGrantReturnPermission = False
        self.lastSentOwnerEntry = 0 # the last time we sent them our IDRecord
        self.lastSentTheirPermission = 0 # last time we sent .theirPermission
        self.watchers = []

    def needToSendTheirPermission(self, debug=False):
        p = self.theirPermission
        if not p:
            if self.lastSentTheirPermission:
                # revoked
                if debug: print "permission was revoked, must update"
                return True
            if debug: print "no permission to update"
        else:
            if debug:
                print "last sent %d, lastModified %d" % \
                      (self.lastSentTheirPermission, p.lastModified)
            if self.lastSentTheirPermission < p.lastModified:
                if debug: print "permission changed recently, must update"
                return True
            if debug: print "permission is up-to-date"
        return False
    def sentTheirPermission(self, when):
        if self.theirPermission:
            self.lastSentTheirPermission = when
        else:
            self.lastSentTheirPermission = 0

    def __getstate__(self):
        d = self.__dict__.copy()
        d['watchers'] = []
        return d

    def notifyOnChange(self, callback):
        """Register a callback to be run when the entry changes (such as
        when the permission is modified, perhaps due to receipt of a message
        from the recipient sender). The callback is run with one argument,
        this AddressBookEntry instance. The callback is not persisted."""
        self.watchers.append(callback)
    def dontNotifyOnChange(self, callback):
        self.watchers.remove(callback)
    def notifyWatchers(self):
        for w in self.watchers:
            w(self)

    def setPetName(self, name):
        self.petname = name

    def createFromIDRecord(self, signed, cagent, cause):
        """Create our .idRecord with the signed IDRecord they just sent us.
        Return a Deferred that will fire (with this entry as an argument)
        when the record is complete."""
        # TODO: record 'cause' as the cause of the update
        assert not self.idRecord
        self.idRecord = IDRecord()
        d = self.idRecord.addFromSigned(signed, cagent)
        d.addCallback(lambda x, entry=self: entry)
        return d

    def updateIDRecord(self, signed, cause):
        """Update this entry with the signed IDRecord they just sent us.
        Return a Deferred that will fire (with this entry as an argument)
        when the update has completed."""
        # TODO: record 'cause' as the cause of the update
        d = self.idRecord.updateFromSigned(signed)
        d.addCallback(self._updated)
        return d

    def _updated(self, results):
        self.notifyWatchers()
        return self

    def updateOurPermission(self, encoded):
        """Accept an encoded permission slip (usually from a
        'yourPermission' message node) and update the .ourPermission
        attribute with its contents.

        You are responsible for calling .modified() after this.
        """
        if not encoded:
            self.ourPermission = None
        else:
            if not self.ourPermission:
                self.ourPermission = permission.PermissionSlip()
            self.ourPermission.setFromEncoded(encoded)

    def _setOurPermission(self, perm):
        # for testing
        assert implements(perm, IPermissionSlip)
        self.ourPermission = perm

    def setTheirPermission(self, perm):
        """You are responsible for calling .modified() after this."""
        assert implements(perm, IOwnerPermission)
        self.theirPermission = perm
    def revokeTheirPermission(self):
        """You are responsible for calling .modified() after this."""
        self.theirPermission = None

    def modified(self):
        # tell the agent that this recipient needs to be refreshed at some
        # point
        self.notifyWatchers()
        if not self.addressbook:
            return
        if not self.doUpdate:
            return
        owner = self.addressbook.owner
        agent = owner.parent
        why = "Must update recipient %s with their new settings" % \
              self.getName()
        d = agent.scheduleEvent(self.updateHoldoffTime, why)
        d.addCallback(self.refreshRecipient)

    def refreshRecipient(self, dummy=None):
        """Returns a Deferred which fires when the update message has been
        sent."""

        if not self.addressbook:
            return # this entry was deleted
        owner = self.addressbook.owner
        if not owner.parent:
            return # the owner was deleted
        agent = owner.parent
        m = Message(agent)
        m.setSenderEntry(owner)
        m.setRecipientEntry(self)
        d = agent.sendMessage(m)
        return d
        

    def getName(self):
        if self.petname != None:
            return self.petname
        if self.idRecord and self.idRecord.name != None:
            return '"%s"' % self.idRecord.name
        if self.idRecord:
            return self.idRecord.ID
        return "????"

    def getTransport(self, agent, message):
        """Obtain a Transport object which can be used to deliver an
        encapsulated message to this recipient.
        """
        t = transport.transportFromString(self.idRecord.transport, agent)
        return t

    def encodePermissionSlip(self):
        """Return a string with an encoded copy of .theirPermission, which
        can be given to the sender. Returns None if they do not have
        specific permission.
        """

        if self.theirPermission:
            return self.theirPermission.encode()
        return None


class FakeNameServer(object):
    def __init__(self):
        self.names = {}
        self.IDs = {}
    def publish(self, record, signed):
        assert implements(record, IOwnerEntry)
        for address in record.addresses:
            self.names[address] = signed
        self.IDs[record.ID] = signed

    def lookup(self, agent, name):
        return defer.succeed(self.names.get(name))
    def lookupID(self, agent, ID):
        return defer.succeed(self.IDs.get(ID))
    def describe(self):
        return "Fake"

class DiskNameServer(object):
    def __init__(self, args):
        self.mapdir = args

    def _getAllRecords(self, agent, records):
        mapdir = os.path.expanduser(self.mapdir)
        if not os.path.exists(mapdir):
            os.mkdir(mapdir)
        defs = []
        for f in os.listdir(mapdir):
            s = open(os.path.join(mapdir, f), "r").read()
            r = IDRecord()
            d = r.createFromSigned(s, agent.crypto)
            d.addCallback(self._gotARecord, records)
            d.addErrback(self._notGotARecord, f)
            defs.append(d)
        dl = defer.DeferredList(defs)
        return dl

    def _gotARecord(self, record, records):
        records.append(record)
    def _notGotARecord(self, failure, f):
        log.err(failure)
        return None

    def lookup(self, agent, address):
        """Take a name and return a (single) ID Record, using a server of
        some sort."""
        records = []
        d = self._getAllRecords(agent, records)
        d.addCallback(self._doLookup, records, address)
        return d
    def _doLookup(self, results, records, address):
        for r in records:
            if address in r.addresses:
                return r.signed
        return None

    def lookupID(self, agent, ID):
        """Take an ID number and return a (single) ID Record, using a server
        of some sort."""
        records = []
        d = self._getAllRecords(agent, records)
        d.addCallback(self._doLookup, records, ID)
        return d
    def _doLookupID(self, results, records, ID):
        for r in records:
            if ID == r.ID:
                return r.signed
        return None

    def publish(self, record, signed):
        assert implements(record, IOwnerEntry)
        mapdir = os.path.expanduser(self.mapdir)
        if not os.path.exists(mapdir):
            os.mkdir(mapdir)
        f = open(os.path.join(mapdir, record.ID), "w")
        f.write(signed)
        f.close()

    def describe(self):
        return "Disk:%s" % self.mapdir

class PBNameServer(object):
    def __init__(self, url):
        self.hostname, portnum = url.split(":")
        self.portnum = int(portnum)

    def lookup(self, agent, address):
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._lookup2, "lookup", address)
        return d

    def lookupID(self, agent, ID):
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._lookup2, "lookupID", ID)
        return d

    def _lookup2(self, server, method, address):
        d = server.callRemote(method, address)
        return d

    def publish(self, record, signed):
        assert implements(record, IOwnerEntry)
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._publish2, signed)
        return d
    def _publish2(self, server, signed):
        d = server.callRemote("publish", signed)
        return d

    def describe(self):
        return "PB:%s:%d" % (self.hostname, self.portnum)    

def getLookupAgentFromString(args):
    scheme, rest = args.split(":", 1)
    fac = {"Fake": FakeNameServer,
           "Disk": DiskNameServer,
           "PB": PBNameServer,
           }
    return fac[scheme](rest)
