#! /usr/bin/python

from twisted.internet import defer

from petmail import permission, util, transport, interfaces

class IDRecord(object):
    """This record describes a recipient. Each recipient in the agent's
    addressbook has one. Each is uniquely identified by the hash of the
    public key.

    The owner of the agent has some OwnerEntry records, and these are
    published in various ways. When somebody looks up one of those published
    records, the data they retrieve is turned into an IDRecord.
    
    Serialized forms of these records are retrieved in the process of
    sending a message. This is the form in which they are stored in the
    addressbook and the 'me' records.
    """

    __implements__ = interfaces.IIDRecord

    def __init__(self):
        self.ID = None     # hash of the public key
        self.pubkey = None # public key
        # identifying information (all optional)
        self.name = ""
        self.addresses = [] # email addresses which point to this record
        # permission record
        self.genericPermission = None # non-sender-specific permissions
        # transport record
        self.transportString = None
        # other stuff
        self.expires = None
        self.signed = None

    def updateGenericPermission(self, encoded):
        if not self.genericPermission:
            self.genericPermission = permission.PermissionSlip()
        self.genericPermission.setFromEncoded(encoded)

    def createFromSigned(self, signed, ca):
        """Take a signed/encoded form (the output of OwnerEntry.sign) and
        populate this record. Do not add the key to the crypto agent. Do not
        try to verify the consistency of the record (like it being signed by
        the public key it provides). The .pubkey attribute will be
        unpopulated, but ID will be present."""
        assert signed
        d = ca.decrypt(signed)
        d.addCallback(self._createFromSigned2, signed)
        return d

    def _createFromSigned2(self, results, signed):
        recip, sigID, encoded = results
        # sigID will probably be None, as this is a new key
        record = util.decode(encoded)
        self.ID = record['ID']
        self.pubkey = None
        self.signed = signed
        self._update(record)
        return self

    def addFromSigned(self, signed, ca):
        """Take a signed/encoded form (the output of OwnerEntry.sign) and
        populate this record."""
        assert signed
        d = ca.decrypt(signed)
        d.addCallback(self._addFromSigned2, signed, ca)
        return d

    def _addFromSigned2(self, results, signed, ca):
        recip, sigID, encoded = results
        # sigID will probably be None, as this is a new key
        record = util.decode(encoded)
        d = ca.addPubkey(record['pubkey'])
        d.addCallback(self._addFromSigned3, signed, ca, record)
        return d

    def _addFromSigned3(self, pubkey, signed, ca, record):
        # now that the key has been added to the keyring, verify it again
        d = ca.decrypt(signed)
        d.addCallback(self._addFromSigned4, pubkey, signed, record)
        return d

    def _addFromSigned4(self, results, pubkey, signed, record):
        recip, sigID, dummy = results
        if pubkey.getID() != sigID:
            raise ValueError("IDRecord signed by the wrong key:" \
                             "key is %s, sig is %s" % (pubkey.getID(),
                                                       sigID))
        if record['ID'] != sigID:
            raise ValueError("IDRecord holds ID field with wrong key:" \
                             "ID field is %s, sig is %s" % (record['ID'],
                                                            sigID))
        self.ID = record['ID']
        self.pubkey = pubkey
        self.signed = signed # to send to others
        self._update(record)
        return self

    def updateFromSigned(self, signed):
        ca = self.pubkey.agent
        d = ca.decrypt(signed)
        d.addCallback(self._updateFromSigned2, signed)
        return d

    def _updateFromSigned2(self, results, signed):
        ca = self.pubkey.agent
        recip, sigID, encoded = results
        if sigID != self.ID:
            raise ValueError("update not correctly signed")
        record = util.decode(encoded)
        if record['ID'] != self.ID:
            raise NotImplementedError("pubkey update")
        d = ca.addPubkey(record['pubkey'])
        d.addCallback(self.updateFromSigned3, signed, record)
        return d

    def updateFromSigned3(self, newpubkey, signed, record):
        if newpubkey.getID() != self.ID:
            # TODO: handle key changes
            raise NotImplementedError("pubkey update")
        self.signed = signed # to send to others
        self._update(record)
        return self

    def _update(self, d):
        self.name = d['name']
        self.addresses = d['addresses']
        self.expires = d['expires']
        slip = d.get('permission')
        if slip:
            self.updateGenericPermission(slip)
        transports = d['transports']
        self.transportString = None
        if transports:
            self.transportString = transports[0] # TODO

    def getTransport(self, agent, message):
        tport = transport.transportFromString(self.transportString, agent)
        return defer.succeed(tport)
