#! /usr/bin/python

from twisted.python.components import implements
from twisted.internet import defer
from twisted.python import failure, log

from petmail import util, crypto, interfaces
from petmail.idrecord import IDRecord
NotPermitted = interfaces.NotPermitted

class InvalidMessage(Exception):
    """The message was mangled somehow"""

class BaseMessage(object):
    logPriority = 0
    debug = False
    def __init__(self, agent):
        self.agent = agent
        # delivery attributes
        self.when = util.now()  # the nominal creation time of the message
        self.eventlog = [] # log of what processors have done
        self.doDump = True
        self.provides = []
        # recipientEntry, recipientRecord, grantedPermission
        # senderEntry, acceptedRFC822
        self.senderEntry = None

    def startProcessing(self):
        if self.agent.debugProcessing:
            print "startProcessing"
        self.processors = self.getProcessors()
        self.processing = []
        self.done = False
        self.failed = False
        d = self.processingDeferred = defer.Deferred()
        self.process()
        return d
        
    def process(self):
        """Do whatever can be done to the message."""
        # this function will not raise Unsendable exceptions. Instead, it
        # (or a callback) will run self.processingFailed
        debug = self.agent.debugProcessing
        if debug:
            text = "process:\n"
            text += " processors %s\n" % self.processors
            text += " provides %s\n" % self.provides
            text += " processing %s\n" % self.processing
            self.log(text)
        if not self.processors and not self.processing:
            if debug:
                self.log("all done")
            return self.processingComplete()
        runnable = self.findRunnableProcesses()
        if not runnable and not self.processing:
            text = "no processors are runnable"
            text += " (processors: %s)" % self.processors
            text += " (provides: %s)" % ",".join(self.provides)
            text += " (processing: %s)" % self.processing
            if debug:
                self.log(text)
                print " no processors are runnable"
            f = failure.Failure
            why = f(interfaces.Unsendable(text))
            return self.processingFailed(why)

        if debug and not runnable:
            self.log("weird, nothing to do")
            print " weird, nothing to do"
        self.processing.extend(runnable)
        for p in runnable:
            lognode = self.log(p.__class__.__name__)
            d = defer.maybeDeferred(p.checkMessage, self, lognode)
            if self.debug:
                self.log("%s gave %s" % (p, d))
            d.addCallback(self.processorComplete, p)
            d.addErrback(self.processorFailed, p)
        if debug:
            self.log(" done")

    def findRunnableProcesses(self):
        # which processors can be run? Removes them from self.processors
        runnable = []
        need = []
        for p in self.processors[:]:
            canRun = True
            for d in p.getDependencies():
                if not d in self.provides:
                    if d not in need: need.append(d)
                    canRun = False
                    break
            needToRun = False
            # empty provides means it must always be run
            processorProvides = p.getProvides()
            if not processorProvides:
                needToRun = True
            else:
                for f in processorProvides:
                    if not f in self.provides:
                        needToRun = True
            if not needToRun:
                if self.agent.debugProcessing:
                    print "don't need to run", p
                self.processors.remove(p)
            elif canRun:
                self.processors.remove(p)
                runnable.append(p)
        if self.agent.debugProcessing:
            self.log("runnable: %s" % ",".join(map(str, runnable)))
            print " runnable:", runnable
        return runnable

    def processorComplete(self, res, processor):
        if self.agent.debugProcessing:
            self.log("processorComplete", processor)
            print "processorComplete", processor
        self.processing.remove(processor)
        if not self.failed:
            self.process()
    def processorFailed(self, why, processor):
        if self.agent.debugProcessing:
            self.log("processorFailed: %s" % processor)
            self.log("failure:", why)
            print "processorFailed", processor, why
        self.processing.remove(processor)
        self.processingFailed(why)

    def processingComplete(self):
        # this should only be called once, when the last processor succeeds.
        assert not self.done
        self.done = True
        self.processingDeferred.callback(self)
    def processingFailed(self, why):
        # this may be called multiple times, once per failing processor
        log.msg("processing step failed:", why)
        if self.failed:
            if self.agent.debugProcessing:
                print "processingFailed >1 times"
            return
        self.done = self.failed = True
        self.processingDeferred.errback(why)

    def provide(self, which):
        if which not in self.provides:
            self.provides.append(which)

    def hasNode(self, name):
        return self.headers.has_key(name)
    def getNode(self, name, default=None):
        return self.headers.get(name, default)

    def startLogging(self, msg):
        if hasattr(self.agent, "eventLog"):
            p = self.agent.eventLog.add(msg, pclass=self.logPriority)
        else:
            p = None
        self.logParent = p

    def log(self, *args, **kwargs):
        if not hasattr(self.agent, "eventLog"):
            return
        if kwargs.has_key('parent'):
            return self.agent.eventLog.add(*args, **kwargs)
        if self.logParent:
            return self.agent.eventLog.add(parent=self.logParent, *args)
        elif hasattr(self.agent, "eventLog"):
            e = self.agent.eventLog.add("late log for %s" % self)
            self.agent.eventLog.add(parent=e, *args)
            return e

    def stopLogging(self):
        if self.logParent:
            self.agent.eventLog.add("logging stopped, message complete",
                                    parent=self.logParent)
            self.logParent = None


class Message(BaseMessage):
    """This object represents a single message being sent to some individual
    recipient. Its life-cycle is as follows:

     1: Created when the user hits the 'Send' button on their MUA,
        containing an rfc822-formatted header+body pair.
     2: The recipient's AddressBookEntry is looked up.
     3: .encode() is called with the AddressBookEntry. This returns a string
        which should be given to the transport layer for delivery to the
        recipient.
     4: The Message is stored in a list of outstanding messages (indexed by
        messageID). An ack from the far end will remove it from the list. If
        it stays on the list for too long (5 days), the user is notified
        that the message did not make it through.


    Useful attributes of this object:

    .senderEntry: the OwnerEntry which is sending the message

    .recipientAddress: set iff the message is being sent to a new recipient for
                       which we only have an address, else None.

    .recipientEntry: at the start of processing, this is only set if the
                     user is writing to a correspondent in their
                     addressbook. After the AddressLookup processor is run,
                     it will be set (and the entry added to the user's
                     AddressBook) if an rfc822 message is being sent to
                     them.

    .recipientRecord: Set as a destination (with .setRecipientRecord), or
                      when AddressLookup turns .recipientAddress into an
                      idRecord. This is what we actually need to deliver the
                      message.
    
    .transport: an ITransport object capable of sending this message

    .provides is a list of things that have been done to this message:

    senderEntry: .senderEntry has been populated # TODO: require this
    recipient: .recipientRecord and .recipientEntry have been populated
    grantedPermission: return permission has been granted

    """
    __implements__ = interfaces.IMessage

    def __init__(self, agent):
        BaseMessage.__init__(self, agent)
        self.messageID = util.generateMessageID()
        self.recipientAddress = None
        self.recipientRecord = None # 'recipientRecord'
        self.recipientEntry = None # 'recipientEntry'
        self.transport = None # the transport used to send the message
        self.worthSending = False
        self.sent = None # the time this message was sent. TODO: set it
        self.headers = {} # indexed by name
        self.references = []
        self.startLogging("Outbound Message created")


    def setMessageRecipient(self, name, owner):
        # this can accept a pet name 
        book = self.agent.getAddressBook(owner.ID)
        r = book.lookupPetName(name)
        # and addresses
        if not r:
            r = book.lookupAddress(name)
        if r:
            self.setRecipientEntry(r)
        else:
            self.recipientAddress = name # will be looked up later

    def setRecipientRecord(self, r):
        assert implements(r, interfaces.IIDRecord)
        self.recipientRecord = r
        self.addNode("recipientID", r.ID)
        self.provide("recipient")

    def setRecipientEntry(self, e):
        assert implements(e, interfaces.IAddressBookEntry)
        self.recipientEntry = e
        self.setRecipientRecord(e.idRecord)
        self.provide("recipient")

    def setSenderEntry(self, e):
        assert implements(e, interfaces.IOwnerEntry), e
        self.senderEntry = e
        self.provide('senderEntry')

    def getProcessors(self):
        return self.agent.getOutboundProcessors()

    def logOutboundMessage(self):
        assert self.senderEntry
        if self.senderEntry:
            self.log(".senderEntry is %s" % self.senderEntry.getName())
        if self.recipientEntry:
            self.log(".recipientEntry is %s" % self.recipientEntry.getName())
        elif self.recipientRecord:
            self.log(".recipientRecord is %s" % self.recipientRecord.name)
        elif self.recipientAddress:
            self.log(".recipientAddress is %s" % self.recipientAddress)
        else:
            self.log("I wonder who this is going to")

    def startProcessing(self):
        self.logOutboundMessage()
        return BaseMessage.startProcessing(self)

    def addNode(self, name, value):
        assert (name not in self.headers)
        self.headers[name] = value

    def addRFC822(self, body): #, us, recipient, body, addrbook=None):
        """Add an RFC822 message section to the Message."""

        # construct the header
        header = {}
        #header['bodyHash'] = util.hash(body)
        #header['bodyLen'] = len(body)
        ##header['senderID'] = us.ID
        ##header['messageID'] = util.generateMessageID()
        ##header['timestamp'] = self.when
        header['tickets'] = []
        header['body'] = body

        self.addNode("rfc822", header)
        self.worthSending = True

    def addReference(self, entry):
        """Add a reference to the given AddressBookEntry. This allows the
        recipient to add the referenced Identity to their own address
        book."""

        assert implements(entry, interfaces.IAddressBookEntry)
        self.references.append(entry)

    def send(self):
        return self.agent.sendMessage(self)

    def addTicket(self, ticket):
        # ticket is a signed blob
        assert(self.headers['rfc822'])
        self.headers['rfc822']['tickets'].append(ticket)

    def encodeHeaders(self):
        return util.encode(self.headers)

    def deliver(self):
        self.log("Delivering message")
        self.log("headers: %s" % self.headers.keys())
        d = self.encrypt(self.senderEntry.privkey,
                         self.recipientRecord.pubkey)
        d.addCallback(self.transport.send)
        return d

    def encrypt(self, us, recipient):
        """Returns a Deferred with the encrypted form of this message.
        If 'us' is present, sign the message.
        If 'recipient' is present, encrypt the message.
        """

        if us:
            if not isinstance(us, crypto.PrivateKey):
                print "us", us
            assert isinstance(us, crypto.PrivateKey)
            self.log("signing with %s" % us.getID())

        if recipient:
            assert isinstance(recipient, crypto.PublicKey)
            self.log("encrypting to %s" % recipient.getID())
        else:
            self.log("not encrypting")
        encoded = util.encode(self.headers)
        if recipient:
            return recipient.encryptAndSign(us, encoded)
        elif us:
            return us.sign(encoded)
        else:
            return self.agent.store(encoded)

    def dump(self):
        if not self.doDump:
            return
        log = self.agent.log
        log("Outbound message dump:")
        log(" worthSending", self.worthSending)
        keys = self.headers.keys()
        keys.sort()
        log("", ",".join(keys))
        for m in self.eventlog:
            log(" %s" % m)
        log(" message dump finished")


class ReceivedMessage(BaseMessage):
    """A ReceivedMessage object is created for each incoming message.

    The .process method is run on it to check permissions. Once this
    succeeds, the following attributes will be available on the
    ReceivedMessage object:

    TODO: add the permission with which the message was accepted

    Useful attributes of this object which are set early:
    .agent: the agent which is receiving the message
    .encrypted: a file object (or StringIO) with the original encrypted form
    .recipientEntry: the OwnerEntry that is the target of this message
    .headers: the unserialized contents of the message

    Attributes which are set by various processing steps (and the .provides
    tags which indicate they are available):

    .allowRFC822 [checkedRFC822]: a boolean indicating that we are willing
    to accept an rfc822 message from this ReceivedMessage

    .senderEntry: always present, but only non-None for messages which have
    a valid signature from a previously-known sender. It may be set later
    (after InboundProcessing is complete) if the user chooses to add the
    sender to their address book.
    
    """

    __implements__ = interfaces.IReceivedMessage

    def __init__(self, agent, encrypted):
        """If any part of the message is invalid, this will raise
        InvalidMessage or UndecryptableError. If the sender does not have
        permission, then it will raise NotPermitted."""

        BaseMessage.__init__(self, agent)
        self.encrypted = encrypted
        self.startLogging("Inbound Message created")

    def decrypt(self):
        """I decrypt my crypttext and set the following attributes:

        .recipientEntry: the OwnerEntry of the message recipient (to whom
        the message was encrypted, or the 'recipientID' node if the message
        was not encrypted)

        .headers: the plaintext message dictionary

        .senderEntry: this will point to an AddressBookEntry if the message
        has a valid signature from a known sender (someone who is already in
        our AddressBook). It will be None if the message had no signature or
        was signed by a sender we don't already know (such as a new sender).

        .senderIDRecord: if .senderEntry is None, the AcceptRFC822
        processing step uses .createIDRecord() to populate this with the
        unverified IDRecord of the otherwise-unknown sender. It is only
        really useful to extract a name for presentation to the user.

        I return a Deferred which will fire with the .recipientEntry .
        """

        ca = self.agent.crypto
        d = ca.decrypt(self.encrypted)
        d.addCallback(self._decrypt2)
        return d

    def _decrypt2(self, results):
        recipientKey, senderID, plaintext = results

        self.headers = util.decode(plaintext)
        self.log("headers: %s" % self.headers.keys())

        try:
            if recipientKey:
                recipientID = recipientKey.getID()
            else:
                recipientID = self.getNode("recipientID")
            self.recipientEntry = self.agent.getSender(recipientID)
        except KeyError:
            self.log("message was meant for someone else")
            raise

        self.senderIDRecord = None
        self.senderEntry = None
        if senderID:
            book = self.agent.getAddressBook(recipientID)
            self.senderEntry = book.lookupID(senderID)
            if self.senderEntry:
                self.senderIDRecord = self.senderEntry.idRecord

        if self.senderEntry:
            who = self.senderEntry.getName()
            self.log("message sent by known sender %s" % who)
        else:
            self.log("message sent by unknown sender")

        return self.recipientEntry

    def getProcessors(self):
        return self.agent.getInboundProcessors()

    def dump(self):
        if not self.doDump:
            return
        log = self.agent.log
        log("Inbound message dump:")
        if self.senderEntry:
            log(" From:", self.senderEntry.getName())
        if "checkedRFC822" in self.provides and self.allowRFC822 and \
            self.getNode("rfc822"):
            l = len(self.getNode('rfc822')['body'])
            log(" RFC822 len(%d) accepted" % l)
        for m in self.eventlog:
            log(" %s" % m)
        log(" message dump finished")

    def getIDRecord(self):
        if self.senderIDRecord:
            return defer.succeed(self.senderIDRecord)
        else:
            return self.createIDRecord()

    def createIDRecord(self):
        signed = self.getNode("senderIDRecord")
        if not signed:
            self.senderIDRecord = None
            return defer.succeed(None)
        r = IDRecord()
        d = r.createFromSigned(signed, self.agent.crypto)
        d.addCallback(self._createIDRecord)
        return d
    def _createIDRecord(self, record):
        self.senderIDRecord = record
        return record

    # canAddSender, canGrantInboundPermission, and canReply are only meant
    # to be called on rfc822-accepted messages

    def canAddSender(self):
        """I check to see if the sender of this message can be added to the
        recipient's AddressBook (i.e.: they are not anonymous, they are not
        already listed). If not, I return True to indicate that the user
        should be offered the option of adding them. To actually add them,
        just call my addSender method."""
        
        if self.senderEntry:
            return False # already there

        book = self.agent.getAddressBook(self.recipientEntry.ID)
        if not book:
            return False # nowhere to put it
        if not self.getNode("senderIDRecord"):
            return False # they didn't give us enough information to add one
        return True

    def addSender(self, reason=None):
        # returns a Deferred that fires with the entry that was added
        self.log("asked to add sender entry")
        book = self.agent.getAddressBook(self.recipientEntry.ID)
        signed = self.getNode("senderIDRecord")
        assert signed
        if not reason:
            reason = self
        d = book.addFromSigned(signed, reason)
        d.addCallback(self._addedSender1)
        return d

    def _addedSender1(self, entry):
        self.senderEntry = entry
        self.log("added sender entry")
        self.grantInboundPermission()
        self.log("granted them inbound permission")
        # now we must update our outbound permission
        slip = self.getNode("yourPermission")
        if not slip:
            self.senderEntry.modified()
            return entry # no specific permission
        d = defer.maybeDeferred(self.senderEntry.updateOurPermission, slip)
        d.addCallback(self._addedSender2, entry)
        return d

    def _addedSender2(self, results, entry):
        self.senderEntry.modified()
        self.log("added our permission")
        return entry

    def canAddReferences(self):
        """I check to see if there are any References in the message
        (pointers to other people that we don't already know about), so the
        user can be offered the option of adding them. I return a list of
        refnodes, one per new Reference. This list might be empty.

        Each refnode is a dictionary. It has a ['name'] node which indicates
        how the sender refers to them, and it may be useful to display this
        to the user. To add the reference, give the refnode to my
        .addReference method.
        """

        if not hasattr(self, 'references'):
            return []
        book = self.agent.getAddressBook(self.recipientEntry.ID)
        newnodes = [r for r in self.references if not book.lookupID(r['id'])]
        return newnodes

    def addReference(self, refnode):
        """I add a reference to our addressbook. If the sender can help us
        get Permission to the referenced recipient, we will sent a request to
        start the process.

        I return a Deferred which will fire when the AddressBookEntry has
        been added. It fires with that entry. Any permission-requesting is
        performed asynchronously.
        """

        lognode = self.log("asked to addReference for %s (%s)" % \
                           (refnode['name'], refnode['id']))
        book = self.agent.getAddressBook(self.recipientEntry.ID)
        reason = self
        d = book.addFromSigned(refnode['idrecord'], reason)
        d.addCallback(self._addedReference1, refnode, lognode)
        return d

    def _addedReference1(self, entry, refnode, lognode):
        self.log("added reference", parent=lognode)
        if refnode['offerPermission']:
            self.log("can get Permission to %s" % entry.getName(),
                     parent=lognode)
            # send permission request
            req = Message(self.agent)
            req.setSenderEntry(self.recipientEntry)
            req.setRecipientEntry(entry)
            request = { 'voucherID': self.senderIDRecord.ID }
            req.addNode("request-permission", request)
            # TODO: make sure the message includes our IDRecord
            # send the request and
            # TODO: tell the agent that we await the response
            d = self.agent.sendMessage(req)
            # ignore the Deferred it returns and any errors caused
            d.addErrback(lambda why: None)
        return entry
        

    def canGrantInboundPermission(self):
        """I check to see if the sender of the message already has inbound
        permission. If not, I return True to indicate that the user should
        be offered the option of granting it to them. To actually grant it,
        just call my grantInboundPermission method.
        """
        entry = self.senderEntry
        if not entry:
            return False # must know who they are to grant them permission
        if entry.theirPermission:
            return False # they already have it
        if entry.doNotGrantReturnPermission:
            return False # we don't want to give it to them
        return True

    def grantInboundPermission(self, reason=None):
        if not reason:
            reason = self
        self.recipientEntry.grantPermissionTo(self.senderEntry, reason)

    def getFromName(self):
        """I return something reasonable to put in the 'from' box. This is
        the user-specified pet name, or the sender-specified name (in
        quotes), or their ID number as a last resort. If there was no sender
        at all (the message was sent anonymously) I will return
        '-anonymous-'."""

        if self.senderEntry:
            return self.senderEntry.getName()
        if self.senderIDRecord:
            if self.senderIDRecord.name:
                return '"%s"' % self.senderIDRecord.name
            return self.senderIDRecord.ID
        return "-anonymous-" # TODO: maybe None to indicate anonymous?

    def canReply(self):
        if self.senderEntry:
            return True
        return False

