#! /usr/bin/python

"""
Define several built-in Processors to handle messages.
"""

"""
Message Processing (outbound):

Messages are created and populated with nodes, then handed to the Agent. The
Agent gives them a list of Processors that get to run (stored in
.processors).

Each processor has .dependencies and .provides lists to provide a sequence
in which they want to be run. Any given processor should be skipped if
everything it .provides is already present in the message. A processor
should not be run until its .dependencies are all provided by the message.

Each Processor can do one of three things (this might be merged into just
one thing later):

 return True, indicating that the processor has done its job immediately
 return a Deferred, which will callback or errback when the job is done
 raise Unsendable

If the Processor returns a Deferred, it is added to the message's
.processing list. When the Deferred completes, it will be removed from that
list.

Message.process figures out all Processors which can be run and starts them.
It

When the message begins processing, the agent puts in on a .processing list. 
"""


from twisted.internet import defer

from petmail import interfaces, util
from petmail.interfaces import Unsendable, NotPermitted
from petmail.message import Message
from petmail.idrecord import IDRecord

class BaseProcessor(object):
    __implements__ = interfaces.IProcessor
    dependencies = []
    provides = []
    def getDependencies(self):
        return self.dependencies
    def getProvides(self):
        return self.provides
    def configureSender(self, sender):
        raise Unimplemented
    def configureRecipient(self, recipient):
        raise Unimplemented
    def assertDependencies(self, message):
        for d in self.dependencies:
            if not d in message.provides:
                raise AssertionError("'%s' not in .provides=%s" \
                                     % (d, message.provides))

class OutboundProcessor(BaseProcessor):
    __implements__ = interfaces.IOutboundProcessor

class InboundProcessor(BaseProcessor):
    __implements__ = interfaces.IInboundProcessor



class Delivery(OutboundProcessor):
    def checkMessage(self, message, lognode=None):
        """Make sure the headers have a MessageID node."""
        self.assertDependencies(message)

        if not message.hasNode("messageID"):
            message.addNode("messageID", message.messageID)
        return True

class AddressLookup(OutboundProcessor):
    provides = ["recipient"]

    def done(self, message):
        message.provide("recipient")
        return True

    def checkMessage(self, message, lognode=None):
        """If the message has a recipientAddress but not a recipientRecord,
        do an address lookup to acquire an IDRecord. Add the recipient's key
        to our keyring (so we can encrypt to them). If we are sending an
        rfc822 message to them, also create an AddressBookEntry. (if not,
        this may result in a key that is not referenced by an
        AddressBookEntry)."""
        self.assertDependencies(message)

        if message.recipientEntry:
            message.log(".recipientEntry present, no lookup needed",
                        parent=lognode)
            return self.done(message) # we have everything
        if message.recipientRecord:
            message.log("record but no entry.. what could get here?",
                        parent=lognode)
            message.agent.log("record but no entry.. what could get here?")
            # we leave message.recipientEntry=None
            return self.done(message)
        if message.recipientAddress:
            d = message.agent.retrieveAddress(message.recipientAddress)
            d.addCallback(self._gotRecord, message, lognode)
            return d
        raise Unsendable, "no address to look up"

    def _gotRecord(self, signed, message, lognode):
        message.log("AddressLookup succeeded", parent=lognode)
        if not signed:
            raise Unsendable("unable to look up address")
        if message.getNode('rfc822'):
            # add the recipient to our address book. This will add their
            # public key as a side-effect.
            if not message.senderEntry:
                # could just return, but I want to know when this might
                # happen
                raise Unimplemented("when can this happen?")
            message.log("adding entry to addressbook", parent=lognode)
            d = self.addEntry(signed, message)
        else:
            # just add their key so we can send the message. This implies
            # that we don't particularly expect to send them another message.
            # TODO: control messages to frequent destinations (remailer users
            # talking to Address/Permission Servers, etc) should reference an
            # IDRecord and not an address, otherwise they'll have to do
            # multiple long lookup cycles for the same target.
            message.log("adding their IDRecord to our cryptoagent",
                        parent=lognode)
            d = self.addRecord(signed, message)
        return d


    def addRecord(self, signed, message):
        record = IDRecord()
        d = record.addFromSigned(signed, message.agent.crypto)
        d.addCallback(self._addedRecord, message)
        return d

    def _addedRecord(self, record, message):
        message.setRecipientRecord(record)
        return self.done(message)


    def addEntry(self, signed, message):
        message.log("installing outbound AddressBook Entry")
        book = message.agent.getAddressBook(message.senderEntry.ID)
        reason = message
        d = book.addFromSigned(signed, reason)
        d.addCallback(self._addedEntry, message)
        return d

    def _addedEntry(self, entry, message):
        message.setRecipientEntry(entry)
        return self.done(message)


class AddReferences(OutboundProcessor):
    dependencies = ["recipient"]

    def checkMessage(self, message, lognode=None):
        """Take any references to other Identities and create the appropriate
        message nodes so the recipient can add them to their addressbook.
        """
        self.assertDependencies(message)

        if not message.references:
            return True # nothing to add
        nodes = []
        for entry in message.references:
            refnode = { 'name': entry.getName(),
                        'id': entry.idRecord.ID,
                        'idrecord' : entry.idRecord.signed,
                        'offerPermission' : 0,
                        }
            lognode2 = message.log("adding %s" % entry.getName(),
                                   parent=lognode)
            # if we vouch for the person we're sending this to,
            #  AND the referencee allows us to grant third-party permission
            # then we tell the recipient that we can help them get permission
            recip = message.recipientEntry
            if recip:
                message.log("vouchFor=%s, ourPermission=%s" % \
                            (recip.vouchFor, bool(entry.ourPermission)),
                            parent=lognode2)
                if entry.ourPermission:
                    message.log("ourPermission.canGrantPermission=%s" % \
                                entry.ourPermission.canGrantPermission,
                                parent=lognode2)
                if recip.vouchFor and entry.ourPermission.canGrantPermission:
                    message.log("offerPermission=1", parent=lognode2)
                    refnode['offerPermission'] = 1
            nodes.append(refnode)
        message.addNode('references', nodes)
        return True


class Transport(OutboundProcessor):
    dependencies = ["recipient"]
    def checkMessage(self, message, lognode=None):
        """If the message does not have a Transport, try to create one."""
        self.assertDependencies(message)

        if message.transport:
            return True
        if not message.recipientRecord:
            raise Unsendable, "cannot get a Transport without an IDRecord"
        #import pdb; pdb.Pdb().set_trace()
        message.log("getting a transport", parent=lognode)
        d = message.recipientRecord.getTransport(message.agent, message)
        if d:
            d.addCallback(self.gotTransport, message, lognode)
        return d
    def gotTransport(self, transport, message, lognode):
        message.log("added transport %s" % transport, parent=lognode)
        message.transport = transport
        return True


class UpdateTheirIDRecord(OutboundProcessor):
    dependencies = ['senderEntry', 'recipient']

    def checkMessage(self, message, lognode=None):
        """If their copy of my ID Record is out of date, include a copy. If
        we don't have specific permission, then we probably aren't in their
        address book, so include a copy regardless of how recently we sent
        the last one."""
        self.assertDependencies(message)

        if not message.recipientEntry:
            # we aren't remembering them, so they probably don't yet know
            # about us. Send it.
            message.log("no recipientEntry, so sending it", parent=lognode)
            return self.addIDRecord(message, lognode)
        if not message.recipientEntry.ourPermission:
            # we probably aren't in their addressbook, so send it.
            message.log("no ourPermission, so sending it", parent=lognode)
            return self.addIDRecord(message, lognode)
        lastSent = message.recipientEntry.lastSentOwnerEntry
        lastModified = message.senderEntry.lastModified
        message.log("lastSent %s, lastModified %s" % (lastSent,
                                                      lastModified),
                    parent=lognode)
        if lastSent < lastModified:
            return self.addIDRecord(message, lognode)
        return True

    def addIDRecord(self, message, lognode):
        d = message.senderEntry.sign()
        d.addCallback(self.gotSigned, message, lognode)
        return d

    def gotSigned(self, signed, message, lognode):
        message.addNode("senderIDRecord", signed)
        message.worthSending = True
        if message.recipientEntry:
            message.recipientEntry.lastSentOwnerEntry = util.now()
        # TODO: update lastSent when ack is received
        message.log("included copy of our IDRecord", parent=lognode)
        return True

class GrantReturnPermission(OutboundProcessor):
    dependencies = ['recipient']
    provides = ['grantedPermission']

    def checkMessage(self, message, lognode=None):
        """When sending an rfc822 message, it is polite to give the
        recipient permission to respond."""
        self.assertDependencies(message)

        # by the time this processor is done, they'll have whatever
        # permission they're going to get
        message.provide("grantedPermission")

        if not message.hasNode("rfc822"):
            return True
        owner = message.senderEntry
        if not owner:
            return
        e = message.recipientEntry
        if not e:
            return True # we can't give them permission if we aren't going
                        # to bother remembering who they are
        if e.theirPermission:
            return True # they already have permission
        if e.doNotGrantReturnPermission:
            message.log("not granting return permission: we don't like them",
                        parent=lognode)
            return True
        if not owner.grantReturnPermission:
            message.log("not granting return permission: we don't do that",
                        parent=lognode)
            return True
        message.log("granting return permission", parent=lognode)
        owner.grantPermissionTo(e, message)
        return True

class UpdateTheirPermission(OutboundProcessor):
    dependencies = ["recipient", "grantedPermission"]

    def checkMessage(self, message, lognode=None):
        """If the recipient has Permission (the 'theirPermission' field of
        their address book entry), and the timestamps indicate that their
        copy is out of date, I will insert a PermissionSlip into the
        outbound message."""
        self.assertDependencies(message)

        if not message.recipientEntry:
            if message.agent.debugProcessing:
                print "no recipientEntry, so no permission to update"
            return True # no permission to be granted
        e = message.recipientEntry
        if not e.needToSendTheirPermission(message.agent.debugProcessing):
            return True
        message.log("updating their permission", parent=lognode)
        message.addNode("yourPermission", e.encodePermissionSlip())
        message.worthSending = True
        e.sentTheirPermission(util.now())
        # TODO: update p.lastSent when ack is received. Retransmits.
        if message.agent.debugProcessing:
            print "did update"
        return True

class RFC822Permission(OutboundProcessor):
    dependencies = ['recipient']
    def checkMessage(self, message, lognode=None):
        """Outbound rfc822 messages require Permission."""
        self.assertDependencies(message)

        node = message.getNode("rfc822")
        if not node:
            return True
        if not message.recipientRecord:
            raise Unsendable, "need ID Record to determine Permission"
        e = message.recipientEntry
        # do we need any kind of permission for this message?
        p = None
        if e and e.ourPermission:
            p = e.ourPermission
            ptype = "(specific)"
            message.log("using specific permission, IMG %s" % p.minIMG,
                        parent=lognode)
            if message.agent.debugProcessing:
                print "Using Specific Permission, IMG", p.minIMG
        else:
            p = message.recipientRecord.genericPermission
            ptype = "(generic)"
            message.log("using generic permission", parent=lognode)
            if message.agent.debugProcessing:
                print "Using Generic Permission", p
        if not p:
            message.log("no permission, rejected", parent=lognode)
            message.log("rfc822 rejected: no permission")
            raise Unsendable, "this recipient does not accept new senders"
        try:
            reqs = p.checkOutboundMessage(message)
        except NotPermitted, why:
            message.log("permission check failed", why, parent=lognode)
            raise
        assert type(reqs) == type([])
        # checkMsg will raise a NotPermitted exception if a size limit is
        # violated. It will return requirements if there are tickets we must
        # obtain before sending
        if not reqs:
            p.sendOutboundMessage(message)
            message.log(ptype + " rfc822 allowed through: no tickets needed",
                        parent=lognode)
            return True
        # we need to fulfill this requirement
        message.log("must fulfill requirements", reqs, parent=lognode)
        if message.agent.debugProcessing:
            print "have reqs", reqs
        dl = []
        tas = message.agent.getTicketAgents()
        if not tas:
            raise Unsendable, ptype + \
                  " message needs tickets but we have no agents to get them"
        for ta in tas:
            deferreds = ta.handleRequirements(reqs, message)
            for d in deferreds:
                if isinstance(d, defer.Deferred):
                    d.addCallback(self.gotTicket, message, p, lognode)
                    d.addErrback(self.failed, message)
                    dl.append(d)
        # TODO: make sure the requirements were actually handled. This
        # probably involves having each ticketAgent tell us which
        # requirements they took care of (or started taking care of)
        if dl:
            # when all these Deferreds fires, the message is ready to go
            dlist = defer.DeferredList(dl, fireOnOneErrback=True)
            # the DeferredList is taking ownership of errors in the
            # individual Deferreds. To make sure no errors get left over,
            # add an errback which will drop them.
            [d.addErrback(lambda why: None) for d in dl]
            return dlist
        return True

    def gotTicket(self, ticket, message, p, lognode):
        message.log("adding ticket", parent=lognode)
        if message.agent.debugProcessing:
            print "gotTicket", ticket
        node = message.getNode('rfc822')
        if not node.has_key("tickets"):
            node["tickets"] = []
        node["tickets"].append(ticket) # TODO: maybe serialize here?
        p.sendOutboundMessage(message)
        return True
    def failed(self, why, message):
        return why


allOutbound = [Delivery, AddressLookup,
               AddReferences,
               Transport, UpdateTheirIDRecord,
               GrantReturnPermission, 
               UpdateTheirPermission, RFC822Permission,
               ]

############################################################

# .senderEntry is looked up first, but set to None if they are unknown

class HandleReply(InboundProcessor):
    def checkMessage(self, message, lognode=None):
        """See if the message is in response to an outstanding request."""
        self.assertDependencies(message)

        replyid = message.getNode("replyID")
        if replyid == None:
            return True
        message.log("gotReply", parent=lognode)
        message.agent.gotReply(replyid, message)
        return True


class HandlePing(InboundProcessor):
    def checkMessage(self, message, lognode=None):
        """See if the message is a ping request."""
        self.assertDependencies(message)

        pingID = message.getNode("ping")
        if pingID:
            message.log("sending ping response", parent=lognode)
            pong = Message(message.agent)
            pong.addNode('pong', pingID)
            message.agent.sendReply(pong, message)
        return True


class CheckRFC822Permission(InboundProcessor):
    provides = ["checkedRFC822"]

    def done(self, message):
        message.provide("checkedRFC822")
        return

    def checkMessage(self, message, lognode=None):
        """Incoming rfc822 messages require valid permission."""
        self.assertDependencies(message)

        a = message.agent
        debug = a.debugProcessing

        message.allowRFC822 = False

        if not message.getNode("rfc822"):
            return self.done(message)
        # they use genericPermission if:
        #  A: we don't know who they are
        #  B: we know them but haven't given them specific permission
        if message.senderEntry and message.senderEntry.theirPermission:
            p = message.senderEntry.theirPermission
            ptype = "(specific)"
            message.log("they have specific permission", parent=lognode)
        else:
            p = message.recipientEntry.genericPermission
            ptype = "(generic)"
            if p:
                message.log("they have generic permission", parent=lognode)
        if not p:
            if debug:
                print " rejected, no permission available"
            message.log("rfc822 rejected: no permission",
                        parent=lognode)
            message.log("rejected rfc822 message: no permission")
            return self.reject(message, NotPermitted("no permission"),
                               lognode)

        if debug:
            print " checking", p
        d = defer.maybeDeferred(p.checkInboundMessage, message)
        d.addCallback(self.didCheck, ptype, p, message, lognode)
        d.addErrback(self.checkFailed, message, lognode)
        return d

    def didCheck(self, why, ptype, p, message, lognode):
        debug = message.agent.debugProcessing
        why = "%s %s" % (ptype, why)
        # if we get here, the message is accepted
        if debug:
            print "  accepted"

        # TODO: do this now? or when message is delivered?
        p.acceptInboundMessage(message)

        message.log("will accept rfc822 message: %s" % why, parent=lognode)
        message.allowRFC822 = why

        # other pieces of the message are handled by other processors
        return self.done(message)

    def checkFailed(self, f, message, lognode):
        f.trap(NotPermitted)
        why = f.value
        return self.reject(message, why, lognode)

    def reject(self, message, why, lognode):
        debug = message.agent.debugProcessing
        if debug:
            print "  rejected", why
        message.log("rejected rfc822 message: %s" % why, parent=lognode)
        message.allowRFC822 = False
        message.agent.rejectRFC822(message, why)
        return self.done(message)


class UpdateMyIDRecord(InboundProcessor):
    provides = ["updatedIDRecord"]

    def done(self, message):
        message.provide("updatedIDRecord")
        return

    def checkMessage(self, message, lognode=None):
        """Update my copy of their ID Record. This is the catcher for the
        update inserted by the UpdateTheirIDRecord outbound processor."""
        self.assertDependencies(message)

        if not message.senderEntry:
            return self.done(message)
        signed = message.getNode("senderIDRecord")
        if not signed:
            return self.done(message)
        # TODO: it would be nice to avoid the decrypt here for cases where
        # we don't need to update anything. Maybe the record's timestamp
        # should be sent outside the signed blob, if it matches our current
        # copy we can skip this step. Doubly useful because updateIDRecord
        # also does an addPubkey.
        message.log("updating my ID Record", parent=lognode)
        d = message.senderEntry.updateIDRecord(signed, message)
        d.addCallback(self._didUpdate, message)
        return d

    def _didUpdate(self, entry, message):
        message.senderIDRecord = message.senderEntry.idRecord
        return self.done(message)

class UpdateOurPermission(InboundProcessor):
    dependencies = ["updatedIDRecord"]
    provides = ["updatedOurPermission"]

    def done(self, message):
        message.provide("updatedOurPermission")
        return

    def checkMessage(self, message, lognode=None):
        """If the inbound message has a PermissionSlip, verify the signature
        and update the OurPermission field of their address book entry. If
        requested, acknowledge receipt of the PermissionSlip. This is the
        catcher for the UpdateTheirPermission outbound processor.
        """
        self.assertDependencies(message)

        if not message.senderEntry:
            return self.done(message) # nothing to update
        if not message.hasNode("yourPermission"):
            return self.done(message)
        slip = message.getNode("yourPermission") # TODO: part of rfc822 node?
        # slip might be None, indicating we have no specific permission
        if slip:
            message.log("updating my PermissionSlip", parent=lognode)
        else:
            message.log("ourPermission was revoked", parent=lognode)
        d = defer.maybeDeferred(message.senderEntry.updateOurPermission,
                                slip)
        d.addCallback(self._didUpdate, message)
        # TODO: send ack, if requested
        return d

    def _didUpdate(self, result, message):
        message.senderEntry.modified()
        return self.done(message)


class AcceptRFC822(InboundProcessor):
    dependencies = ["checkedRFC822",
                    "updatedIDRecord", "updatedOurPermission"]

    def checkMessage(self, message, lognode=None):
        """Finally accept an incoming rfc822 message."""
        self.assertDependencies(message)

        a = message.agent
        debug = a.debugProcessing

        if not message.getNode("rfc822"):
            return
        if not message.allowRFC822:
            return

        # try to create some kind of sender information
        if not message.senderEntry:
            message.log("accepting rfc822 message, but extracting IDRecord"
                        " for sender name first", parent=lognode)
            d = message.createIDRecord()
        else:
            message.log("accepting rfc822 message", parent=lognode)
            d = defer.succeed(None)
        d.addCallback(self._haveSender, message)
        return d

    def _haveSender(self, results, message):
        # tell the agent to deliver the body to the user. At this point,
        # message.senderName() will work.
        message.agent.acceptRFC822(message, message.allowRFC822)
        return True

class HandleReferences(InboundProcessor):
    def checkMessage(self, message, lognode=None):
        self.assertDependencies(message)

        refs = message.getNode("references")
        if not refs:
            return True
        references = []
        for r in refs:
            message.log("reference: offer=%d %s %s" % (r['offerPermission'],
                                                       r['name'],
                                                       r['id']),
                        parent=lognode)
            references.append(r)
        message.references = references
        return True

class HandlePermissionRequest(InboundProcessor):
    def checkMessage(self, message, lognode=None):
        self.assertDependencies(message)

        req = message.getNode("request-permission")
        if not req:
            return True
        # message.senderIDRecord wants permission to send to us.
        voucherID = req['voucherID']
        # they claim that this person vouches for them
        book = message.agent.getAddressBook(message.recipientEntry.ID)
        voucher = book.lookupID(voucherID)
        if not voucher:
            return True
        if not voucher.theirPermission:
            return True
        if not voucher.theirPermission.canGrantPermission:
            return True
        # we trust this voucher to recommend senders. Send them a query to
        # see if they really do vouch for this person; if we get back a
        # signed message stating that they do, we'll accept them.
        
        # we need the requestor's ID, which probably hasn't been extracted
        # yet
        if message.senderIDRecord:
            d = defer.succeed(message.senderIDRecord)
        else:
            d = message.createIDRecord()
        d.addCallback(self._gotIDRecord, message, voucher, lognode)
        return d

    def _gotIDRecord(self, record, message, voucher, lognode):
        requestorID = record.ID

        req = Message(message.agent)
        req.setSenderEntry(message.recipientEntry)
        req.setRecipientEntry(voucher)
        request = { 'requestorID': requestorID }
        req.addNode("check-voucher", request)

        d = message.agent.sendRequest(req)
        # slow: this response will take a roundtrip to the voucher
        d.addCallback(self._gotVouchResponse,
                      requestorID, voucher, message, lognode)
        return d

    def _gotVouchResponse(self, reply,
                          requestorID, voucher, message, lognode):
        owner = message.recipientEntry
        # the response must be from the voucher to whom we sent the request
        if not reply.senderEntry:
            return True # message was unsigned
        if (reply.senderEntry.idRecord.ID != voucher.idRecord.ID):
            return True # message was forged: signed by the wrong party
        replynode = reply.getNode('check-voucher-reply')
        if replynode['requestorID'] != requestorID:
            return True # weird, they answered the wrong question
        if not replynode['doVouch']:
            return True # they said no

        # ok, we are now willing to grant them permission. We have to
        # add them to the addressbook first
        reason = reply
        d = message.addSender(reason)
        d.addCallback(self._added, message, lognode)
        return d

    def _added(self, entry, message, lognode):
        reason = "recommended"
        message.grantInboundPermission(reason)
        return True


class HandleVoucherRequest(InboundProcessor):
    def checkMessage(self, message, lognode=None):
        self.assertDependencies(message)

        req = message.getNode("check-voucher")
        if not req:
            return True
        requestorID = req['requestorID']
        # requestorID claims that we vouch for them
        doVouch = False
        
        book = message.agent.getAddressBook(message.recipientEntry.ID)
        requestor = book.lookupID(requestorID)
        if requestor:
            doVouch = requestor.vouchFor

        reply = Message(message.agent)
        reply.addNode('check-voucher-reply', { 'requestorID': requestorID,
                                               'doVouch': doVouch })
        message.agent.sendReply(reply, message)


allInbound = [HandleReply, HandlePing,
              CheckRFC822Permission,
              UpdateMyIDRecord,
              UpdateOurPermission,
              AcceptRFC822,
              HandleReferences,
              HandlePermissionRequest,
              HandleVoucherRequest,
              ]

# outbound:
#  make sure there is a MessageID header node
#  require that the message has a recipientRecord
#  if the message has a recipientAddress but not a recipientRecord, do an
#   address lookup
#  make sure there is a Transport, get one if not
#  if their copy of the ID Record is out of date, include a copy
#  make sure there is a senderID node (or use the crypto envelope for that?)
#  if their permission record is out of date, include a copy
#  rfc822 messages need permission
#  return permission: if we're sending an rfc822 node to them, and they don't
#   have return permission, and the address book entry doesn't specifically
#   say we should deny them permission, then add Permission to their entry
#  3rd-party permission: if the message references other addresses to whom we
#   have permission, ask the user and then request 3rd-party permission for
#   them (keep track of the answer in the address book entry)

# inbound:
#  if they sent us an rfc822 message that was accepted, make sure there is
#   an address book entry for them
#  if they included an ID Record, make sure our copy is up to date (maybe ack)
#  if they included a Permission Slip, update our copy (maybe ack)
#  if there is an rfc822 node, verify the permission, and deliver to the user
#   (maybe ack)
#  if their message was not deemed undesireable, and they don't already have
#   permission, grant it (probably with a configurable timer)
#  if they requested 3rd-party permission, maybe grant it

# other inbound
# for address servers:
#   respond to addressLookup nodes
# for transport servers:
#   respond to transportLookup nodes
