#! /usr/bin/python

from twisted.trial import unittest
from twisted.internet import reactor, protocol
from twisted.protocols import basic

import os, sys
from StringIO import StringIO
from twisted.python import log
#log.startLogging(sys.stderr)

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import transport, util
from petmail.owner import OwnerEntry

class FakeAgent:
    pass

class TestInlet(unittest.TestCase):

    def testDummy(self):
        i = transport.DummyInlet("foo")
        desc = i.encode()
        self.failUnlessEqual(desc, "Dummy:foo")

    def testPB(self):
        i = transport.PBInlet("localhost", 0)
        desc = i.encode()
        self.failUnlessEqual(desc, "PB:localhost:0")

    def testMailSocket(self):
        i = transport.MailSocket(0, "alice@foo.com")
        desc = i.encode()
        self.failUnlessEqual(desc, "SMTP:alice@foo.com")

    def testMaildir(self):
        i = transport.MaildirInlet("~/mdir-petmail", "alice@foo.com")
        desc = i.encode()
        self.failUnlessEqual(desc, "SMTP:alice@foo.com")
        

class TestTransport(unittest.TestCase):
    def setUp(self):
        self.messages = []

    def receiveMessage(self, msg):
        self.messages.append(msg)

    def testDummy(self):
        i = transport.DummyInlet("foo")
        i.parent = self
        t = transport.transportFromString(i.encode(), None)
        t.send("message")
        self.failUnlessEqual(self.messages, ["message"])

    def testPB(self):
        i = transport.PBInlet("localhost", 0)
        i.parent = self
        i.startService()
        portnum = i.services[0]._port.getHost()[2]
        target = "PB:localhost:%d" % portnum
        t = transport.transportFromString(target, None)
        t.send("message")
        now = util.now()
        while util.now() < now+5:
            if self.messages:
                break
            reactor.iterate(0.1)
        i.stopService()
        self.failUnlessEqual(self.messages, ["message"])

    def testEncapsulate(self):
        message = "message\x00\x01\x02" * 40
        encoded = transport.encapsulateInText(message)
        # the encoded form should survive weird newlines and arbitrary
        # header/footer additions

        encoded.replace("\n", "\n\r", 5)
        encoded = "here to fool you\n" + encoded + "footer\r" + "feet\n"
        encoded = encoded + "little bwitty toeses\n\n"
        
        i = transport.SMTPEncapsulatedInlet()
        i.parent = self

        i.gotText(encoded)
        self.failUnlessEqual(self.messages, [message])

        self.messages = []
        s = StringIO(encoded)
        i.gotFile(s)
        self.failUnlessEqual(self.messages, [message])

    def gotFile(self, fp):
        m = fp.read()
        self.receiveMessage(m)

    def testMaildir(self):
        os.mkdir("maildir")
        newdir = os.path.join("maildir", "new")
        olddir = os.path.join("maildir", "cur")
        tmpdir = os.path.join("maildir", "tmp")
        os.mkdir(newdir)
        os.mkdir(olddir)
        os.mkdir(tmpdir)
        i = transport.MaildirInlet("maildir", "foo")
        i.gotFile = self.gotFile
        i.pollinterval = 0.5
        i.startService()
        self.failUnlessEqual(self.messages, [])
        open(os.path.join(tmpdir, "bar"), "w").write("message")
        os.rename(os.path.join(tmpdir, "bar"), os.path.join(newdir, "bar"))
        now = util.now()
        while util.now() < now+5:
            if self.messages:
                break
            reactor.iterate(0.1)
        i.stopService()
        self.failUnlessEqual(self.messages, ["message"])
        self.failIf(os.path.exists(os.path.join(newdir, "bar")))
        self.failUnless(os.path.exists(os.path.join(olddir, "bar")))

    def testMail(self):
        if not os.environ.get("PETMAIL_TEST_SEND"):
            # this actually sends mail, only do it manually
            raise unittest.SkipTest("define PETMAIL_TEST_SEND=1 to run this")
        a = FakeAgent
        a.smtpScheme = "Mail:localhost:me@localhost"
        i = transport.MaildirInlet("~/Maildir-petmailtest", "foo")
        i.parent = self
        i.pollinterval = 0.5
        i.startService()
        ts = transport.transportFromString
        t = ts("SMTP:warner-petmailtest@localhost", a)
        message = "message\x00\x01\x02" * 40
        d = t.send(message)
        r = unittest.deferredResult(d)
        now = util.now()
        while util.now() < now+5:
            if self.messages:
                break
            reactor.iterate(0.1)
        i.stopService()
        self.failUnlessEqual(self.messages, [message])

    def testMailSocket(self):
        i = transport.MailSocket(0, "alice@foo.com")
        i.parent = self
        i.startService()
        portnum = i.services[0]._port.getHost()[2]
        # this is nicked from twisted.test.test_tpfile.FileSendingTestCase

        class FileSendingClient(protocol.Protocol):
            def __init__(self, f):
                self.f = f

            def connectionMade(self):
                s = basic.FileSender()
                d = s.beginFileTransfer(self.f, self.transport, lambda x: x)
                d.addCallback(lambda r: self.transport.loseConnection())

        class SendingFactory(protocol.ClientFactory):
            def buildProtocol(self, addr):
                return self.p

        message = "message\x00\x01\x02" * 40
        e = transport.encapsulateInText(message)
        f = SendingFactory()
        f.p = FileSendingClient(StringIO(e))
        reactor.connectTCP("localhost", portnum, f)
        now = util.now()
        while util.now() < now+5:
            if self.messages:
                break
            reactor.iterate(0.1)
        i.stopService()
        self.failUnlessEqual(self.messages, [message])
        
