#! /usr/bin/python

"""
This file contains various forms of Transport objects, which are used to get
an encapsulated message body (a binary string) to their receiver. The most
likely choice is a MIME-encoded rfc822 message send via the legacy SMTP
infrastructure, but anything which reliably gets the blob to the other end
will work.

The Transport is also responsible for doing an end-to-end acknowledgement,
retransmitting if necessary.
"""

import os, os.path
import email
from email import Encoders, Parser, Message
import base64

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from twisted.python import log
from twisted.spread import pb
from twisted.cred import portal
from twisted.application import service, internet
from twisted.internet import protocol, error, defer, reactor

from twisted.protocols.smtp import sendmail
from twisted.protocols.pop3 import POP3Client
from twisted.protocols import imap4
from twisted.mail import relaymanager

from petmail import pbutil, interfaces

# the recipient adds an Inlet to their agent
# the Inlet produces a transport string (.encode)
# the transport string is published in the ID Record
# the sender gets the ID Record and extracts the transport string
# the transport string has a prefix and the args
# the prefix is looked up in knownTransports and used to create a Transport

knownTransports = {}
def registerTransport(prefix, factory):
    knownTransports[prefix] = factory

def transportFromString(string, agent):
    if not string:
        return Null(None, agent)
    colon = string.index(":")
    type = string[:colon]
    args = string[colon+1:]
    transport = knownTransports[type](args, agent)
    transport.string = string # TODO? why?
    return transport

class BaseInlet(service.MultiService):
    __implements__ = (interfaces.IInlet, service.MultiService.__implements__)

    def encode(self):
        return self.description
    def __cmp__(self, them):
        c = cmp(self.__class__, them.__class__)
        if c: return c
        return cmp(self.__dict__, them.__dict__)
    def gotMessage(self, encoded):
        self.parent.receiveMessage(encoded)


dummyInlets = {}
class DummyInlet(BaseInlet):
    """DummyInlet/DummyTransport are paired like PTY devices. Both live in
    the same process, and writing to one causes the message to be emitted on
    the other. Used only for testing.

    Dummy:name
    """

    def __init__(self, name):
        BaseInlet.__init__(self)
        self.name = name
        dummyInlets[self.name] = self
        self.description = "Dummy:%s" % self.name
    def accept(self, encrypted):
        return self.parent.receiveMessage(encrypted)

class DummyTransport:
    __implements__ = interfaces.ITransport,
    def __init__(self, args, agent):
        self.name = args
    def send(self, encapsulated):
        inlet = dummyInlets[self.name]
        return inlet.accept(encapsulated)
registerTransport("Dummy", DummyTransport)


class Null:
    __implements__ = interfaces.ITransport,

    def __init__(self, args, agent):
        pass
    def send(self, encapsulated):
        print "discarding message"
registerTransport("Null", Null)



class PBInlet(BaseInlet, pb.Root):
    __implements__ = (BaseInlet.__implements__, pb.Root.__implements__)

    def __init__(self, hostname, portnum):
        BaseInlet.__init__(self)
        self.description = "PB:%s:%d" % (hostname, portnum)
        self.hostname = hostname
        self.portnum = int(portnum)
        server = pbutil.listen(self, int(portnum))
        server.setServiceParent(self)

    def __cmp__(self, them):
        c = cmp(self.__class__, them.__class__)
        if c: return c
        return cmp((self.hostname, self.portnum),
                   (them.hostname, them.portnum))

    def remote_message(self, encrypted):
        return self.parent.receiveMessage(encrypted)
        
    def startService(self):
        BaseInlet.startService(self)

    def stopService(self):
        BaseInlet.stopService(self)
        # probably need to wait a moment here if the socket is going to
        # be re-opened right away


class PBTransport:
    __implements__ = interfaces.ITransport,
    def __init__(self, args, agent):
        host, portnum = args.split(":")
        self.host = host
        self.portnum = int(portnum)
    def send(self, encapsulated):
        d = pbutil.connect(self.host, self.portnum)
        d.addCallback(self._connected, encapsulated)
        d.addErrback(log.err)
        return d

    def _connected(self, ref, encapsulated):
        d = ref.callRemote("message", encapsulated)
        d.addCallback(self._sent, ref)
        d.addErrback(log.err)
        return d
    def _sent(self, result, ref):
        ref.broker.transport.loseConnection()
    def __repr__(self):
        return "petmail.transport.PB<%s:%d>" % (self.host, self.portnum)
registerTransport("PB", PBTransport)


SMTP_SUBJECT = "PETMAIL MESSAGE"
SMTP_HEADER = "-----BEGIN PETMAIL MESSAGE-----"
SMTP_FOOTER = "-----END PETMAIL MESSAGE-----"

class SMTPEncapsulatedInlet(BaseInlet):

    # the next three methods may raise an exception, in which case the error
    # should be logged and the message dropped (or perhaps put in a log
    # directory for a limited period of time)

    def gotText(self, text):
        """This method should be called when the inlet has received an
        incoming message ready to be scanned for the ---BEGIN PETMAIL.. line
        This particular function should be given the message as a string."""
        f = StringIO(text)
        self.gotFile(f)

    def gotFile(self, fp):
        """This function receives the message as an open file. The object
        must support the readline() method.
        """
        while 1:
            line = fp.readline()
            if not line:
                # EOF too early
                raise ValueError("Malformed message")
            if line.rstrip() == SMTP_HEADER:
                line = fp.readline()
                if not line:
                    raise ValueError("Malformed message")
                break
        body = ""
        while 1:
            line = fp.readline()
            if not line:
                # EOF too early
                raise ValueError("Malformed message")
            if line.strip() == SMTP_FOOTER:
                break
            body += line

        encoded = base64.decodestring(body)
        self.gotMessage(encoded)


class PollingSMTPInlet(SMTPEncapsulatedInlet):
    startedTimer = False
    pollinterval = 600 # seconds

    def startService(self):
        if not self.startedTimer:
            # note that the TimerService is not pickleable until it has been
            # stopped.
            # TODO: consider using the deferred to prevent overlapping checks
            timer = internet.TimerService(self.pollinterval, self.check)
            timer.setServiceParent(self)
            self.startedTimer = True
        SMTPEncapsulatedInlet.startService(self)

    def check(self):
        """I check the mailbox and return a Deferred which will fire when the
        check has completed.
        """
        raise Unimplemented

class MaildirInlet(PollingSMTPInlet):
    """I watch a maildir for inbound messages."""

    pollinterval = 10 # seconds

    def __init__(self, basedir, address):
        PollingSMTPInlet.__init__(self)
        self.basedir = basedir
        self.address = address
        self.newdir = os.path.join(os.path.expanduser(basedir), "new")
        self.olddir = os.path.join(os.path.expanduser(basedir), "cur")
        self.description = "SMTP:%s" % address

    def __cmp__(self, them):
        c = cmp(self.__class__, them.__class__)
        if c: return c
        return cmp((self.basedir, self.address),
                   (them.basedir, them.address))

    def check(self):
        for f in os.listdir(self.newdir):
            newfile = os.path.join(self.newdir, f)
            oldfile = os.path.join(self.olddir, f)
            self.gotFile(open(newfile, "r"))
            os.rename(newfile, oldfile)
                      

class MailSocketProtocol(protocol.Protocol):
    def __init__(self):
        self.data = ""
    def dataReceived(self, data):
        self.data += data
    def connectionLost(self, reason):
        if reason.check(error.ConnectionDone):
            self.factory.parent.gotText(self.data)

class MailSocket(SMTPEncapsulatedInlet):
    # [MailSocket:]9990:alice@foo.com
    # socketnum : inletaddress

    def __init__(self, portnum, address):
        SMTPEncapsulatedInlet.__init__(self)
        assert type(portnum) == int
        self.portnum = portnum
        self.address = address
        f = protocol.ServerFactory()
        f.protocol = MailSocketProtocol
        f.parent = self
        server = internet.TCPServer(portnum, f)
        server.setServiceParent(self)
        self.description = "SMTP:%s" % address

    def __cmp__(self, them):
        c = cmp(self.__class__, them.__class__)
        if c: return c
        return cmp((self.portnum, self.address),
                   (them.portnum, them.address))

# IMAP grabber

class IMAPPollProtocol(imap4.IMAP4Client):
    def serverGreeting(self, caps):
        # caps is a dictionary. caps['AUTH'] is a list of login methods
        d = self.login(self.factory.username, self.factory.password)
        d.addCallback(self.loggedIn)
        d.addCallback(self.factory.complete)
        d.addErrback(self.factory.failed)

    def loggedIn(self, *args):
        if self.factory.doDelete:
            d = self.select(self.factory.folder) # read-write
        else:
            d = self.examine(self.factory.folder)
        d.addCallback(self.examine_cb)
        return d

    def examine_cb(self, info):
        if info['EXISTS']:
            d = self.fetchHeaders("1:*")
            d.addCallback(self.gotHeader)
            return d

    def gotHeader(self, msgs):
        ids = msgs.keys()
        ids.sort()
        wanted = imap4.MessageSet()
        for msgid in ids:
            headertext = msgs[msgid]['RFC822.HEADER']
            m = email.message_from_string(headertext)
            print "%s: %s" % (msgid, m['subject'])
            if m['subject'] == "PETMAIL MESSAGE":
                wanted.add(msgid)
        print "wanted:", wanted
        if wanted:
            d = self.fetchBody(wanted)
            d.addCallback(self.gotBody, wanted)
            return d

    def gotBody(self, msgs, wanted):
        print "gotBody"
        ids = msgs.keys()
        ids.sort()
        for msgid in ids:
            text = msgs[msgid]['RFC822.TEXT']
            self.factory.inlet.gotText(text)
        if self.factory.doDelete:
            print "deleting:", wanted
            # TODO: twisted-1.1.2 will let us remove the str()
            d = self.setFlags(str(wanted), ("\\Deleted",))
            d.addCallback(self.didDelete)
            return d
        else:
            self.close()

    def didDelete(self, *args):
        print "didDelete", args
        self.close()

class IMAPPollerFactory(protocol.ClientFactory):
    protocol = IMAPPollProtocol

    def __init__(self, username, password, folder):
        self.username = username
        self.password = password
        self.folder = folder # might be an empty string
        self.deferred = defer.Deferred()
        self.done = False

    def clientConnectionFailed(self, connector, reason):
        self.failed(reason)

    def clientConnectionLost(self, connector, reason):
        self.failed(reason)

    def complete(self, results):
        print "complete"
        self.done = True
        self.deferred.callback(results)

    def failed(self, why):
        if not self.done:
            self.done = True
            print "failure", why
            self.deferred.errback(why)

class IMAPPoller(PollingSMTPInlet):
    doDelete = True

    def __init__(self, host, portnum, username, password, pollinterval,
                 folder, address):
        PollingSMTPInlet.__init__(self)
        self.host = host
        self.portnum = portnum
        self.username = username
        self.password = password
        self.pollinterval = pollinterval
        self.address = address
        self.folder = folder # might be an empty string
        self.description = "SMTP:%s" % address

    def check(self):
        print "connecting.."
        f = IMAPPollerFactory(self.username, self.password, self.folder)
        f.doDelete = self.doDelete
        f.inlet = self
        reactor.connectTCP(self.host, self.portnum, f)
        return f.deferred



class POPPollProtocol(POP3Client):
    # incomplete
    def handle_WELCOME(self, line):
        POP3Client.handle_WELCOME(self, line)
        self.user(self.factory.poller.username)
    def handle_USER(self, data):
        pass

    
class POPPoller(PollingSMTPInlet):

    def __init__(self, host, portnum, username, password, pollinterval,
                 address):
        raise Unimplemented # incomplete
        SMTPEncapsulatedInlet.__init__(self)
        self.host = host
        self.portnum = portnum
        self.username = username
        self.password = password
        self.pollinterval = pollinterval
        self.address = address
        self.description = "SMTP:%s" % address

    def check(self):
        print "checking"
                      


# SMTP-encapsulating output. The Agent has a .smtpScheme string which defines
# a preference for how we should do outbound SMTP. This scheme is
# "Mail:relay:fromaddr" for simple mailing, and "Mixminion:??" for
# send-through-mixminion

smtpSchemes = {}
def makeSMTPTransport(args, agent):
    scheme, schemeArgs = agent.smtpScheme.split(":", 1)
    transport = smtpSchemes[scheme](args, agent)
    return transport
registerTransport("SMTP", makeSMTPTransport)

def encapsulateInText(encoded):
    """This puts the PETmail message body into a ---PETMAIL MESSAGE---
    payload block. It returns a string ready to be inserted into the body
    of a Message object. The subject line of the resulting message must
    be 'PETMAIL MESSAGE'."""

    text = base64.encodestring(encoded)
    body = "%s\n\n%s%s\n" % (SMTP_HEADER, text, SMTP_FOOTER)
    return body

class JustMailIt(object):
    """This Transport class uses twisted.mail to deliver the message via SMTP
    to a designated relay host. For normally-abled unix boxes that run an MTA
    like sendmail, this defaults to 'localhost'. Others will need to set it
    to a host which *is* capable of sending mail.
    """
    __implements__ = interfaces.ITransport,

    def __init__(self, args, agent):
        scheme, self.smarthost, self.fromEmail = agent.smtpScheme.split(":")
        assert scheme == "Mail"
        self.dest = args

    def send(self, encoded):
        """This returns a Deferred which will fire when the message has been
        sent successfully."""
        m = Message.Message()
        m.set_payload(encapsulateInText(encoded))
        m['Subject'] = SMTP_SUBJECT
        m['X-PETmail-URL'] = "http://petmail.lothar.com/"
        m['From'] = self.fromEmail
        m['To'] = self.dest
        s = m.as_string()
        return sendmail(self.smarthost, self.fromEmail, self.dest, s)

smtpSchemes["Mail"] = JustMailIt

class MailAnonymously(object):
    """This subclass sends the message out through a remailer network, making
    it nigh-impossible to figure out where it came from. The only information
    available to the recipient (or an eavesdropper) is the contents of the
    IDRecord."""

    pass
smtpSchemes["Mixminion"] = MailAnonymously

class UsenetDropbox(object):

    """This subclass posts the message to a Usenet group, probably
    alt.anonymous.messages, where the recipient can pick it up at their
    (fairly anonymous) leisure. The sender is not particularly anonymous
    unless they send the message through a remailer network."""

    pass

# TODO: the SMTP Tranports have two kinds of configurable state: that which
# comes from the sender's configuration, and that which comes from the
# recipient's TransportRecord.

# recipient defines:
#  "I want messages via SMTP at address bob@foo.com"
#  "I want messages to be posted to alt.anonymous.messages with subject
#   '123'"

# sender defines:
#  "All SMTP messages should be sent with a From: address of alice@foo.com"
#  "All SMTP messages should be sent through smarthost smtp.isp.com"
#  "All SMTP messages should be sent through mixminion"
#  "Usenet posts should be posted to nntp.isp.com"
#  "Usenet posts should be mail-to-newsed through mixminion and m2n.blah.com"

# Likewise, the Inlet contains state that is purely local, which should not
# be published in their TransportRecord. The published data is listed above.

# recipient-local data:
#  "I get SMTP mail by sucking POP from user/pass @pop.isp.com every 5min"
#  " or by IMAP"
#  "I get SMTP by listening on port 25"
#  "I get SMTP by sucking on a maildir at ~/Maildir-petmail every 5 minutes"
#  "I get SMTP by watching ~/Maildir-petmail for changes with FAM or such"
#  "I get SMTP by one of the above, and sometimes doing nymserver-IMAP
#    polling when the nymserver tells me there are a lot of messages queued,
#    and oh by the way please keep the nymserver supplied with SURBs"
#  "I get Usenet posts by polling nntp.isp.com every 30 minutes"
#  "I get SMTP by listening on localhost:9990 or the ~/.petmail-inlet UNIX
#    socket, because my .forward/.qmail says '|nc localhost 9990'"
